/**
 * Copyright (C) 2010-2012 Agrosense info@agrosense.eu
 *
 * Licensed under the Eclipse Public License - v 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *         http://www.eclipse.org/legal/epl-v10.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package nl.cloudfarming.client.sensor;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.geom.AffineTransform;
import java.awt.geom.NoninvertibleTransformException;
import javax.swing.JPanel;
import nl.cloudfarming.client.geoviewer.Palette;
import nl.cloudfarming.client.geoviewer.ValueRange;
import org.geotools.coverage.grid.GridCoverage2D;
import org.geotools.coverage.grid.ViewType;
import org.geotools.factory.CommonFactoryFinder;
import org.geotools.filter.expression.ExpressionAbstract;
import org.geotools.referencing.crs.DefaultGeographicCRS;
import org.geotools.renderer.lite.gridcoverage2d.GridCoverageRenderer;
import org.geotools.styling.*;
import org.opengis.filter.expression.ExpressionVisitor;
import org.opengis.referencing.FactoryException;
import org.opengis.referencing.operation.TransformException;
import org.openide.util.Exceptions;

/**
 *
 * @author Timon Veenstra
 */
public class GridCoveragePanel extends JPanel {

    /**
     * this gets rid of exception for not using native acceleration
     */
    static {
        System.setProperty("com.sun.media.jai.disableMediaLib", "true");
    }
    private StyleFactory sf = CommonFactoryFinder.getStyleFactory(null);
    private final SensorGrid grid;
    private Palette palette;

    public GridCoveragePanel(SensorGrid grid, Palette palette) {
        this.grid = grid;
        this.palette = palette;

    }

    @Override
    public void paint(Graphics g) {
        super.paint(g);
        try {
            GridCoverage2D coverage2D = grid.getRenderObject(null);
            double minx = coverage2D.getEnvelope2D().getMinX();
            double maxx = coverage2D.getEnvelope2D().getMaxX();
            double miny = coverage2D.getEnvelope2D().getMinY();
            double maxy = coverage2D.getEnvelope2D().getMaxY();


            com.vividsolutions.jts.geom.Envelope rasterEnvelope = new com.vividsolutions.jts.geom.Envelope(minx, maxx, miny, maxy);
            Rectangle myRect = rectWithCorrectedAspectRatio(getVisibleRect(), maxx - minx, maxy - miny);
            GridCoverageRenderer renderer = new GridCoverageRenderer(DefaultGeographicCRS.WGS84, rasterEnvelope, myRect, (AffineTransform) null);
            RasterSymbolizer symbolizer = sf.getDefaultRasterSymbolizer();
            symbolizer.setOpacity(new OpacityExp(0.9f));

            if (palette instanceof ValueRange) {
                ValueRange range = (ValueRange) palette;
                Number min = range.getMinValue();
                Number max = range.getMaxValue();
                double step = (max.doubleValue() - min.doubleValue()) / range.getSteps();
                for (int s = 0; s < range.getSteps(); s++) {
                    Integer value = new Double(min.doubleValue() + (s * step)).intValue();
                    Color c = palette.getColorForValue(value);
                    symbolizer.getColorMap().addColorMapEntry(getEntry(c, value.intValue()));
                }

            } else {
                //
                //fallback
                //
                symbolizer.getColorMap().addColorMapEntry(getEntry(new Color(0xFF000000, true), 1));
                symbolizer.getColorMap().addColorMapEntry(getEntry(new Color(0xFF8F050C, true), 100));
                symbolizer.getColorMap().addColorMapEntry(getEntry(new Color(0xFFDB0F01, true), 200));
                symbolizer.getColorMap().addColorMapEntry(getEntry(new Color(0xFFE65E00, true), 300));
                symbolizer.getColorMap().addColorMapEntry(getEntry(new Color(0xFFFC9F00, true), 400));
                symbolizer.getColorMap().addColorMapEntry(getEntry(new Color(0xFFFAD007, true), 500));
                symbolizer.getColorMap().addColorMapEntry(getEntry(new Color(0xFFDCF40A, true), 600));
                symbolizer.getColorMap().addColorMapEntry(getEntry(new Color(0xFF92FC00, true), 700));
                symbolizer.getColorMap().addColorMapEntry(getEntry(new Color(0xFF41C405, true), 800));
                symbolizer.getColorMap().addColorMapEntry(getEntry(new Color(0xFF048608, true), 900));
                symbolizer.getColorMap().addColorMapEntry(getEntry(new Color(0xFF044606, true), 1000));
            }
            symbolizer.getColorMap().setType(ColorMap.TYPE_RAMP);

            renderer.paint((Graphics2D) g, coverage2D.view(ViewType.GEOPHYSICS), symbolizer);
        } catch (FactoryException | TransformException | NoninvertibleTransformException ex) {
            Exceptions.printStackTrace(ex);
        }
    }
    /**
     * Dirty correction to maintain a fixed aspect ratio for the displayed geographical sensor values
     */
    private Rectangle rectWithCorrectedAspectRatio(Rectangle rect, double sizeX, double sizeY) {
        int maxx = rect.width;
        int maxy = rect.height;

        // Introduce artificial correction factor to pretend to translate coordinates to pixels correct.
        double xxx = (maxy / sizeY) * (sizeX * 0.7);
        double yyy = (maxx / (sizeX * 0.7)) * sizeY;

        double correctedx = maxx;
        double correctedy = maxy;

        if (xxx > maxx) {
            correctedy = yyy;
        } else {
            correctedx = xxx;
        }
        rect.setBounds(0, 0, (int) correctedx, (int) correctedy);
        return rect;
    }

    public void setPalette(SensorPalette sensorpalette) {
        this.palette = sensorpalette;
    }

    private ColorMapEntry getEntry(Color color, int value) {
        ColorMapEntry entry = new ColorMapEntryImpl();
        entry.setColor(new ColorExp(color));
        entry.setQuantity(new ValueExp(value));
        return entry;
    }

    private class ColorExp extends ExpressionAbstract {

        private final Color color;

        public ColorExp(Color color) {
            this.color = color;
        }

        @Override
        public Object evaluate(Object object) {
            return this.color;
        }

        @Override
        public Object accept(ExpressionVisitor visitor, Object extraData) {
            throw new UnsupportedOperationException("Not supported yet.");
        }
    }

    private class ValueExp extends ExpressionAbstract {

        private final int value;

        public ValueExp(int value) {
            this.value = value;
        }

        @Override
        public Object evaluate(Object object) {
            return Integer.valueOf(this.value);
        }

        @Override
        public Object accept(ExpressionVisitor visitor, Object extraData) {
            throw new UnsupportedOperationException("Not supported yet.");
        }
    }

    private class OpacityExp extends ExpressionAbstract {

        private final double opacity;

        public OpacityExp(double opacity) {
            assert opacity <= 1.0 && opacity > 0.0;
            this.opacity = opacity;
        }

        @Override
        public Object evaluate(Object object) {
            return new Double(this.opacity);
        }

        @Override
        public Object accept(ExpressionVisitor visitor, Object extraData) {
            throw new UnsupportedOperationException("Not supported yet.");
        }
    }
}
